import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.colors import rgb2hex
import svgwrite

class Vis:
    padding = 5
    square_s = 50

    def __init__(self, seqs, max_n_color=6, legend=None, mapping_name=None, title=None, width_in_block=20):
        """
        Produce SVG visualizations for event logs
        :param seqs: event logs represented as list of lists
        :param max_n_color: maximum number of colors, only the {max_n_color}
                            most appearing activities will have their own color
        :param legend:  For some analysis, it could be useful to have two event logs that
                        shares the same legend (e.g., when doing clustering); i.e.,
                        two visualizations will share the same colors and are hence comparable.
                        This param allow to apply an existing legend to the graph.
                        The legend is extracted from the graph using self.get_legend().
                        If none, the legend is built automatically by looking at the
                        most occuring activities
        :param mapping_name: In a web service context, we can save some space if we provide an
                             index for the activity name instead of the real name;
                             i.e., [['this is activity 1','activity2']] becomes [[1,2]]
                             mapping_name allows to map the real name to the index so the
                             legend will show the real name.
                             this is how the mapping_name would look like:
                             {1:'this is activity 1', 2:'this is activity 2'}
        :param title:   Optional title at the top of the visualization
        :param width_in_block:  Max number of activities per sequence (default: 20).
                                If a trace contains more than {width_in_block}, it will
                                be truncated like [1,2,3,4,5,...,10] (with width_in_block=7)
        """
        self.seqs = self.load_seqs(seqs)
        self.max_n_color = int(max_n_color)
        self.legend = legend
        self.mapping_name = mapping_name
        self.title = title
        self.width_in_block = width_in_block

        if not self.legend:
            self.legend = self.extract_legend(self.seqs)
        else:
            self._add_missing_activity_to_legend()

        self.svg = {}
        # We create three kind of SVGs
        for type in ['seq','legend','seqlegend']:
            self.svg[type] = self._build_svg(type)


    def _add_missing_activity_to_legend(self):
        '''
        When using an existing legend, we should make sure that all
        activities from the current event logs are also added to the legend
        (with the color 'others')
        :return:
        '''
        last_row = None
        highest_ranking = 0
        for v in self.legend.values():
            if v['ranking']>highest_ranking:
                highest_ranking = v['ranking']
                last_row = v
        for k,v in self.extract_legend(self.seqs).items():
            if k not in self.legend:
                self.legend[k] = last_row
                self.legend[k]['o_name'] = k

    def load_seqs(self, seqs):
        return [[str(y) for y in x] for x in seqs]

    def extract_legend(self, seqs):
        '''
        Assign a color to each activities.
        The top {self.max_n_color-1} activities that occurs the most will
        have a distinct color while the remaining one will get the label 'others'
        :param seqs: event logs as list of lists
        :return: a dictionary
        '''

        # Count activities
        y = pd.Series([y for x in seqs for y in x]).value_counts().to_frame()

        # Prepare dataframe
        y.columns = ['count']
        y.index.name = 'name'
        y = y.reset_index()
        y = y.sort_values(by=['count','name'], ascending=[False, True])

        # Assign color
        n_color = min(self.max_n_color, y.shape[0])
        palette = plt.get_cmap('magma', n_color+1)
        colors = [rgb2hex(palette(x)) for x in range(n_color)]
        y['color'] = colors[-1]
        y.loc[y.iloc[0:n_color].index, 'color'] = colors

        # Map potential name
        mapping = {x:x for x in y['name'].tolist()}
        if self.mapping_name:
            for k,v in self.mapping_name.items():
                if k in mapping.keys():
                    mapping[k] = v
        y.index = y['name'].tolist()
        y['o_name'] = y['name'].copy()
        y['name'] = y['name'].map(mapping)
        y['ranking'] = np.arange(y.shape[0])

        if y.shape[0] > self.max_n_color:
            y.loc[y.iloc[n_color-1:].index, 'name'] = '+ {} others...'.format(y.shape[0]-n_color+1)
            y.loc[y.iloc[n_color-1:].index, 'color'] = '#eee'
        return y.to_dict(orient='index')

    def _build_svg(self, type):
        '''
        Draw the SVG.
        :param type: There are three types of SVGs:
                    1. 'legend': show only the legend
                    2. 'seq': show only the sequence of activities
                    3. 'seqlegend': show both (1) and (2)
        :return: a svgwrite object
        '''

        n = 0
        if type in ['seq','seqlegend']:
            n += len(self.seqs)   # For the sequence
            n += 1              # For the number of cases at the top
        if type in ['legend','seqlegend']:
            # For the legend
            n += min(self.max_n_color, len(self.legend)) + 2

        height = n*(self.padding+self.square_s)
        width = self.width_in_block * self.square_s
        font_size = self.square_s*0.7


        d = svgwrite.Drawing('test.svg', profile='tiny', size=(width, height))

        if type in ['seq','seqlegend']:
            d.add(d.text(self.title,  insert=(5, 25), fill='black', font_size=font_size,))

            for row, trace in enumerate(self.seqs):
                for col, activity in enumerate(trace):
                    top = (((row+(self.title is not None))*self.square_s) + ((row)*self.padding+1))
                    if len(trace) > self.width_in_block:
                        if col == len(trace) - 2:
                            left = (self.width_in_block - 2)*self.square_s
                            r = self.square_s/15
                            d.add(d.circle((left+(self.square_s/2), top+(self.square_s/2)), r, fill='#000000'))
                            d.add(d.circle((left+(self.square_s/2)-(r*3), top+(self.square_s/2)), r, fill='#000000'))
                            d.add(d.circle((left+(self.square_s/2)+(r*3), top+(self.square_s/2)), r, fill='#000000'))
                            continue
                        if col == len(trace)-1:
                            left = (self.width_in_block - 1)*self.square_s
                            d.add(d.rect((left, top), (self.square_s, self.square_s), fill=self.legend[trace[-1]]['color']))
                            continue
                        if col > self.width_in_block - 3:
                            continue
                    left = col*self.square_s
                    d.add(d.rect((left, top), (self.square_s, self.square_s), fill=self.legend[activity]['color']))

        if type in ['legend','seqlegend']:
            top = 1 if type == 'legend' else len(self.seqs) + 2
            for i, v in enumerate(self.legend.values()):
                if v['ranking']>=self.max_n_color:
                    continue
                t = (self.square_s + self.padding)*(v['ranking']+top)
                d.add(d.rect((0, t), (self.square_s*0.8, self.square_s*0.8), fill=v['color']))
                d.add(d.text(v['name'],  insert=(self.square_s, t+(self.square_s*0.6)), fill='black', font_size=font_size,))
        d.viewbox(0, 0, width, height)
        d.fit(horiz='left', vert='top', scale='meet')
        return d

    def get_svg(self, type):
        """
        Return a SVG in text format
        :param type: There are three types of SVGs:
                    1. 'legend': show only the legend
                    2. 'seq': show only the sequence of activities
                    3. 'seqlegend': show both (1) and (2)
        :return: text in xml format
        """
        return self.svg[type].tostring()

    def get_activities(self):
        '''
        Return the legend, useful when we want to apply it to another visualizations.
        :return: dictionary
        '''
        return self.legend

    def save_svg(self, type, path):
        '''
        Save the SVG to the specified path
        :param type: There are three types of SVGs:
                    1. 'legend': show only the legend
                    2. 'seq': show only the sequence of activities
                    3. 'seqlegend': show both (1) and (2)
        :param path: path
        '''
        with open(path, 'w') as f:
            f.write(self.get_svg(type))



import math
import numpy as np
np.seterr(divide='ignore', invalid='ignore')
import pandas as pd
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.decomposition import TruncatedSVD
from py2opt.routefinder import RouteFinder
import random2
from itertools import combinations
import editdistance
import os
import sys
from sklearn.neighbors import NearestNeighbors

class CminSampler:
    def __init__(self, seqs, p=0, heuristic_presampling=10000, tsp_sort=True, random_seed=1):
        """
        Reduce the size of an event logs, while making sure that the selected traces are representatives
        :param seqs: list of list representing the event logs (e.g., [[1,2],[1,2,3],[1,2]]
        :param p: Size of the event logs once sampled. If set to 1, the size is automatically choosen as a function of the original size
        :param heuristic_presampling: Apply a random sampling (proportional to variants' size) to make the process faster.
        :param tsp_sort: If set to true: Sort the sequence by similarity using a TSP algorithm
        :param random_seed: Given the same seed and identical event logs, make sure the sampling, forces the same output
        """
        self.seqs = seqs
        self.variants = pd.Series(self.seqs).value_counts()

        if p<1 or p>self.variants.shape[0]:
            p = self.auto_size()

        self.p = p
        self.heuristic_presampling = heuristic_presampling
        self.tsp_sort = tsp_sort
        self.random_seed = random_seed

        self.multiplicity = self.variants.values
        self.variant_seq = self.variants.index.tolist()

        # n. traces assigned to each rep'
        self.capacity = math.floor(self.multiplicity.sum()/self.p)

        # Expected Occurrence Reduction reduces the number of operations needed (see paper)
        ratio = self.multiplicity.sum()/self.p
        count = self.multiplicity / ratio
        self.eo_counts = np.floor(count) # Add frequently appearing variants
        self.multiplicity = ((count - self.eo_counts) * ratio).round(0).astype(int) # Update the multiplicity

    def auto_size(self):
        """
        Define p based on the number of traces
        """
        return max(int(round(math.log(float(self.variants.sum()),1.5),0)),1)

    def sample(self):
        """
        Select the p most representative traces
        """
        if len(self.seqs) == 1:
            return self.seqs
        sampler = self.samplingWithEucl()
        seqs = [y for i, x in enumerate(sampler) if x > 0 for y in [self.variant_seq[i]]*int(x)]

        if self.tsp_sort and sampler[sampler>0].shape[0]>2:
            sys.stdout = open(os.devnull, 'w')
            seqs = self.tsp_sorting(seqs)
            sys.stdout = sys.__stdout__
        return seqs

    def samplingWithEucl(self):
        """
        Run the sampling algorithm in the Euclidean space
        """

        data = self.buildSignature()
        data = data.repeat(self.multiplicity, axis=0)

        original_seq_index = np.arange(self.multiplicity.shape[0]).repeat(self.multiplicity)
        not_assigned = np.ones(data.shape[0]).astype(bool)

        if self.heuristic_presampling:
            if data.shape[0] > self.heuristic_presampling:
                np.random.seed(self.random_seed)
                r = np.random.choice(not_assigned.shape[0], not_assigned.shape[0] - self.heuristic_presampling, replace=False)
                not_assigned[r] = False
                self.capacity = math.floor(self.heuristic_presampling / self.p)

        output_count = []
        while len(output_count) != self.p - self.eo_counts.sum():
            i_not_assigned = np.where(not_assigned==True)[0]
            ldata = data[not_assigned,:]

            if ldata.shape[0] > 2:
                neigh = NearestNeighbors(n_neighbors=min(self.capacity,ldata.shape[0]))
                neigh.fit(ldata)
                D, I = neigh.kneighbors(ldata, return_distance=True)
                best_id = D.sum(axis=1).argmin()
                closest_to_best = I[best_id,:]

                output_count.append(original_seq_index[i_not_assigned[best_id]])
                not_assigned[i_not_assigned[closest_to_best]] = False
            else:
                # Add random seq (because choosing the most central from less than 3 obs is not possible)
                missing = self.p - len(output_count) - self.eo_counts.sum()
                np.random.seed(self.random_seed)
                output_count.extend(np.random.choice(original_seq_index[i_not_assigned], int(missing)).tolist())

        return np.bincount(output_count, minlength=self.multiplicity.shape[0]) + self.eo_counts

    def buildSignature(self):
        """
        Extract features using ngrams and reduce dimensionality with SVD
        """
        cv = CountVectorizer(ngram_range=(1,2), tokenizer=lambda doc: doc, lowercase=False, max_features=1024)
        seqs = [['$$START$$']+list([str(y) for y in x])+['$$END$$'] for x in list(self.variants.keys())]
        data = cv.fit_transform(seqs)
        data = TruncatedSVD(min(8, int(data.shape[1]/2)+1), random_state=0).fit_transform(data).astype(np.float32)
        return data

    def tsp_sorting(self, seqs):
        '''
        Order the sequence by similarty using a TSP algorithm
        :param seqs: event logs represented as list of lists
        :return:
        '''
        dist_mat = self.buildDistanceMatrix(seqs)
        random2.seed(self.random_seed)
        route_finder = RouteFinder(dist_mat, None, iterations=5)
        best_distance, best_route = route_finder.solve()

        d = []
        for i in range(len(best_route)):
            current = best_route[i]
            try:
                next = best_route[i+1]
            except:
                next = best_route[0]
            d.append(dist_mat[current][next])
        cut = np.array(d).argmax()
        best_route = best_route[cut+1:]+best_route[:cut+1]

        return [seqs[int(x)] for x in best_route]


    def buildDistanceMatrix(self, seq):
        '''
        Build distance matrix between sequences using the normalize edit distance
        :param seq: event logs provided as list of list
        :return: Matrix of size (len(seqs)^2)
        '''
        m = np.zeros([len(seq), len(seq)])
        for x, y in combinations(range(0,len(seq)), 2):
            d = editdistance.eval(seq[x], seq[y]) / max([len(seq[x]), len(seq[y])])
            m[x,y] = d
            m[y,x] = d
        for x in range(len(seq)):
            m[x,x] = 0
        return m.astype(np.float64)
